/*
 * Copyright (c) 2023 The wayvnc authors
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once

#include <stdint.h>
#include <stdbool.h>

struct output;
struct zwlr_output_manager_v1;
struct zwlr_output_configuration_v1;

void wlr_output_manager_setup(struct zwlr_output_manager_v1* output_manager);
void wlr_output_manager_destroy(void);

struct zwlr_output_configuration_v1* wlr_output_manager_start_config(void);
bool wlr_output_manager_commit_config(struct zwlr_output_configuration_v1* config);
void wlr_output_manager_abort_config(struct zwlr_output_configuration_v1* config);
bool wlr_output_manager_configure_output(struct zwlr_output_configuration_v1* config,
	struct output* output, uint16_t width, uint16_t height, uint16_t x,
	uint16_t y);

bool wlr_output_manager_resize_output(struct output* output,
	uint16_t width, uint16_t height);
