/* Generated by wayland-scanner 1.25.0 */

/*
 * Copyright © 2018 Ilia Bozhinov
 * Copyright © 2020 Isaac Freund
 * Copyright © 2022 wb9688
 * Copyright © 2023 i509VCB
 *
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include "wayland-util.h"

#ifndef __has_attribute
# define __has_attribute(x) 0  /* Compatibility with non-clang compilers. */
#endif

#if (__has_attribute(visibility) || defined(__GNUC__) && __GNUC__ >= 4)
#define WL_PRIVATE __attribute__ ((visibility("hidden")))
#else
#define WL_PRIVATE
#endif

extern const struct wl_interface ext_foreign_toplevel_handle_v1_interface;

static const struct wl_interface *ext_foreign_toplevel_list_v1_types[] = {
	NULL,
	&ext_foreign_toplevel_handle_v1_interface,
};

static const struct wl_message ext_foreign_toplevel_list_v1_requests[] = {
	{ "stop", "", ext_foreign_toplevel_list_v1_types + 0 },
	{ "destroy", "", ext_foreign_toplevel_list_v1_types + 0 },
};

static const struct wl_message ext_foreign_toplevel_list_v1_events[] = {
	{ "toplevel", "n", ext_foreign_toplevel_list_v1_types + 1 },
	{ "finished", "", ext_foreign_toplevel_list_v1_types + 0 },
};

WL_PRIVATE const struct wl_interface ext_foreign_toplevel_list_v1_interface = {
	"ext_foreign_toplevel_list_v1", 1,
	2, ext_foreign_toplevel_list_v1_requests,
	2, ext_foreign_toplevel_list_v1_events,
};

static const struct wl_message ext_foreign_toplevel_handle_v1_requests[] = {
	{ "destroy", "", ext_foreign_toplevel_list_v1_types + 0 },
};

static const struct wl_message ext_foreign_toplevel_handle_v1_events[] = {
	{ "closed", "", ext_foreign_toplevel_list_v1_types + 0 },
	{ "done", "", ext_foreign_toplevel_list_v1_types + 0 },
	{ "title", "s", ext_foreign_toplevel_list_v1_types + 0 },
	{ "app_id", "s", ext_foreign_toplevel_list_v1_types + 0 },
	{ "identifier", "s", ext_foreign_toplevel_list_v1_types + 0 },
};

WL_PRIVATE const struct wl_interface ext_foreign_toplevel_handle_v1_interface = {
	"ext_foreign_toplevel_handle_v1", 1,
	1, ext_foreign_toplevel_handle_v1_requests,
	5, ext_foreign_toplevel_handle_v1_events,
};

